function oResultModel = kglvq_2010(iNrOfPrototypes,data,rlen,alpha0,varargin)
% Kernel Generalized Learning Vector Quantizer in accordance to Qin and Suganthan
% Implementation and additional Optimization by Frank-M. Schleif
%
% Optimizations
% - nystroem approximation
%
% oResultModel = kglvq_2010(n,oDataStruct,rlen,alpha,lambda,sigma)
%
%   oResultModel = kglvq_2010(3,oDataStruct,10,0.01,3/2,0.5);
%
%  Input and output arguments: 
%   n       (scalar-int)    number of expected clusters
%   D       (struct)        data struct, the class information must
%                           be present on the first column of .labels field
%                           labels given as cellstr(numstr(numeric_labels))
%                            - labels are not used in KNG but are good for the post labeling -
%                           For precalculated kernel D = N x N gram matrix of the kernel
%   rlen    (scalar-int)    running length
%   alpha0  (scalar-double) learning parameter (annealed), e.g. 0.05
%   lambda0 (scalar-double) neighborhood parameter (annealed), e.g. iNrOfPrototypes/2 - in kernel space
%
%   oResultModel (struct) result model - can be used e.g. with PostLabel
%
% varargin parameters all as 'parameter_name',value or 'parameter_name','value'
% parameter name              value                    description          
%   nystroem                0.01 -- 1.0                specifies the approximation using Nystroem method (~0 high approximation, 1 no approximation)
%   few_memory              0/1                        0 - full kernel usage, 1 - nystroem used
%   kernel                  @rbf,@linear               kernel type, please specify a functor to the implementation
%                                                      e.g. like f = @(x)rbf(x,0.5) with x as a free parameter  
%                                                      with nystroem please use an appropriate implementation
%
% NOTE: does not take mask into account.
%
% For more help, try 'type kglvq_2010', or check out online documentation.
% Frank-Nichael Schleif schleif@informatik.uni-leipzig.de
%
fSparseOls = @(x,sigma0) (log(1+(x/(sigma0+eps)).^2)); 
bUseSparsity = false;
beta0 = 1;
tStart=tic;
pKernelFunc = @linear;
bDebug = 0;
bFewMemory = false;
bResetRandom = false;
iRandNumber  = 1234;
bNystroem    = false; 
dSampleRate = 1.0; % 50%
bActive      = false;
bPreInit = false;
bNormalize = true;
if(~isempty(varargin))
    vMatches = find(cellfun(@numel,arrayfun(@(x) ((strfind(cell2mat(x),'few_memory'))), varargin,'UniformOutput',false))==1);
    if(~isempty(vMatches))
        dTemp    = cell2mat(varargin(vMatches(1)+1));
        if(isnumeric(dTemp) && dTemp == 1)
            bFewMemory = true;
        end
    end
    
    vMatches = find(cellfun(@numel,arrayfun(@(x) ((strfind(cell2mat(x),'nystroem'))), varargin,'UniformOutput',false))==1);
    if(~isempty(vMatches))
        dTemp    = cell2mat(varargin(vMatches(1)+1));
        if(isnumeric(dTemp))
            dSampleRate = dTemp;
            if(dSampleRate > 1)
                dSampleRate = 1.0;
            end
        end
        bNystroem  = true;
        bFewMemory = true;        
    end
    
    vMatches = find(cellfun(@numel,arrayfun(@(x) ((strfind(cell2mat(x),'sparse_beta'))), varargin,'UniformOutput',false))==1);
    if(~isempty(vMatches))
        dTemp    = cell2mat(varargin(vMatches(1)+1));
        if(isnumeric(dTemp))
            beta0 = dTemp;
            if(beta0 < 0)
                beta0 = 0.1;
            elseif(beta0 > 0)
                bUseSparsity = true;
                bNormalize = true;
            end
        end        
    end    
    
    vMatches = find(cellfun(@numel,arrayfun(@(x) ((strfind(cell2mat(x),'random'))), varargin,'UniformOutput',false))==1);
    if(~isempty(vMatches))
        dTemp    = cell2mat(varargin(vMatches(1)+1));
        if(isnumeric(dTemp))
            iRandNumber = ceil(dTemp);
            bResetRandom = true;
            if(iRandNumber < 0)
                iRandNumber  = 1234;
            end
        end
    end

    vMatches = find(cellfun(@numel,arrayfun(@(x) ((strfind(cell2mat(x),'active'))), varargin,'UniformOutput',false))==1);
    if(~isempty(vMatches))
        sTemp    = cell2mat(varargin(vMatches(1)+1));
        if(ischar(sTemp) && strcmp(sTemp,'false'))
            bActive = false;
        else
            bActive = sTemp;
        end
    end    
    
    vMatches = find(cellfun(@numel,arrayfun(@(x) ((strfind(cell2mat(x),'kernel'))), varargin,'UniformOutput',false))==1);
    if(~isempty(vMatches))
        sTemp       = cell2mat(varargin(vMatches(1)+1));        
        if(strcmp(class(sTemp),'function_handle'))
            pKernelFunc = sTemp;
        else
            warning('Invalid function handle for the kernel parameter - use default'); 
        end
    end
    
end

if(bActive)
    bNormalize = true;
end
rand('twister');
bRandStateStored = false;
savedState = [];
if(isempty(strfind(version('-release'),'2007')))
    defaultStream = RandStream.getDefaultStream;
    savedState    = defaultStream.State;
    bRandStateStored = true;
end
if(bResetRandom)
    rand('twister',iRandNumber);
end

bGenerateKernel = false;
if(isstruct(data))
    bGenerateKernel = true;
    dat = data.data;
    iNrOfDataPoints = size(dat,1);
else % assume its a kernel
    K = data; 
    iNrOfDataPoints = size(data,1);
end

d_class = data.labels(:,1);

if(iscell(d_class))
    d_class = str2num(cell2mat(d_class));
end

vUniqueLabels  = unique(d_class);
vDClassNumeric = zeros(length(d_class),1);
for(i=1:length(vUniqueLabels))
    vDClassNumeric(d_class == vUniqueLabels(i)) = i;
end

vCClassNumeric = mod(randperm(iNrOfPrototypes),length(vUniqueLabels))+1;
c_class=vCClassNumeric';
d_class=vDClassNumeric;
eps2     = 1.0/iNrOfDataPoints * 0.01; 


% nyquist approximation 
N        = iNrOfDataPoints;
M        = ceil(N * dSampleRate); % number of projection dimensions left
if(M == 0)
    M = 1;
end
P        = randperm(N);
ind      = P(1:M);

if(~bFewMemory) % use direct calculations
    if(bGenerateKernel)
        K = pKernelFunc(dat);
    end
    nyDis{1} = K(:,ind);         % C
    nyDis{2} = pinv(K(ind,ind)); % W^+_k
else
    [C,W]    = pKernelFunc(dat,ind);
    nyDis{1} = C;
    nyDis{2} = W;
end
nyDis{3} = P;
vDiag    = zeros(iNrOfDataPoints,1);
for(l=1:iNrOfDataPoints)
    vDiag(l) = nyDis{1}(l,:)*nyDis{2} * nyDis{1}(l,:)';
end
% nyDis can be used to calculate distances 
vOnes    = ones(iNrOfDataPoints,1);
vOnes2   = vOnes';
G = (rand(iNrOfPrototypes,iNrOfDataPoints));
% normalize G
Normalizer = sum(G,2);
Normalizer = Normalizer(:,vOnes2);
if(bPreInit)
    for(i=1:length(vUniqueLabels)) % for each label
        G(c_class == i,d_class ~= i) = 0;
    end
end
G = G./Normalizer;
G = sparse(G);

vRandomDataIndices = randperm(iNrOfDataPoints);
vMeanDists = zeros(iNrOfPrototypes,2);
alpha = alpha0 * (0.005/alpha0).^([0:(rlen-1)]/rlen);

iSkipped      = 0;
if(bUseSparsity)
    vSparsityLog = zeros(rlen,2);
else
    vSparsityLog = zeros(1,1);
end
vGNormalized = sum(G,1)./iNrOfPrototypes;
vGStd        = (2*std(vGNormalized).^2)+eps;
dExpWeight   = 1/iNrOfDataPoints;
vThresh      = exp(-(vGNormalized-dExpWeight).^2 ./ (vGStd));
vDistTmp     = sparse(zeros(iNrOfPrototypes,iNrOfDataPoints));  
for t=1:rlen
  fprintf('\rTraining round: %d/%d',t,rlen);   
  if(t==1) % define some starting parameter in the first round 
        vDistTmp = ((G*nyDis{1})*nyDis{2})*nyDis{1}';        
        % following is not updated
        mTemp       = G .* vDistTmp; % diag(G * K * G')'; --- sum(G .* vDistTmp,2) are the distances to the prototypes
        sigmaSparse = (var(mTemp,[],2)); % dSigma2 - thats just a setting - to estimate sigma by the statistic on wF
        sigmaSqP    = 1.0./(sigmaSparse+eps);
  end
  iSkip = 0;
  for m=1:iNrOfDataPoints 
    j = vRandomDataIndices(m);
    
	if(bActive)        
        dThresh = vThresh(j);
        dRand   = rand(1);
        if(0.5*dThresh < dRand)
                iSkipped = iSkipped+1;
                continue;
        end
    end

    %vDistancesToPrototypes = -2 * vDistTmp(:,j) + diag(G * vDistTmp') + vDiag(j);
    vDistancesToPrototypes = -2 * vDistTmp(:,j) + sum(G .* vDistTmp,2) + vDiag(j);    

    if bDebug && min(vDistancesToPrototypes)<0
        warning('at least one distance is negative.');
    end
    vDistancesToPrototypes(vDistancesToPrototypes<0) = 0;

    vIndicesCorr  = find(c_class == d_class(j));
    vIndicesICorr = find(c_class ~= d_class(j));
    [dCorr, iMinCorr]  = min(vDistancesToPrototypes(vIndicesCorr));
    [dICorr,iMinICorr] = min(vDistancesToPrototypes(vIndicesICorr));    
    iMinCorr       = vIndicesCorr(iMinCorr);
    iMinICorr      = vIndicesICorr(iMinICorr);    
    vMeanDists(iMinCorr,1) = vMeanDists(iMinCorr,1) + dCorr;
    vMeanDists(iMinCorr,2) = vMeanDists(iMinCorr,2) + 1;
    if(0 && t > 10 && dCorr < dICorr && dCorr < 1.5 * vMeanDists(iMinCorr,1)/vMeanDists(iMinCorr,2))
        iSkip = iSkip +1;
        continue;
    end
    dFrac          = dCorr+dICorr+eps;
    glvq_mu        = (dCorr-dICorr)./(dFrac);
    sgd_pr         = sigmoid_prime(glvq_mu);
    sgd_pr         = alpha(t)*sgd_pr;        
    dCoeffDist     = sgd_pr * 4 * dICorr./dFrac;
    dCoeffDistI    = sgd_pr * 4 * dCorr./dFrac;

       
    if(bUseSparsity)
        % calc
        mTemp       = G .* vDistTmp; % diag(G * K * G')'; --- sum(G .* vDistTmp,2) are the distances to the prototypes
        %sigmaSparse = (var(mTemp,[],2)); % dSigma2 - thats just a setting - to estimate sigma by the statistic on wF    
        %sigmaSqP    = 1.0./(sigmaSparse+eps);
        mTemp       = bsxfun(@times,1+mTemp,sigmaSqP);
        mDeriv      = 2* beta0 ./ (mTemp+eps); % +1 for numerical stab
        mSparsityUpdate = -bsxfun(@times,bsxfun(@times,G,sigmaSqP),mDeriv); % multiplied by G because this is the rest of wF after elementwise division of phi from the other side of the equation
    end

    vCorr  = G(iMinCorr,:);
    vICorr = G(iMinICorr,:);
    
    G(iMinCorr,:)  = vCorr  -  dCoeffDist  .* vCorr; 
    G(iMinICorr,:) = vICorr +  dCoeffDistI .* vICorr;    
    G(iMinCorr,j)  = vCorr(j) + dCoeffDist ;
    G(iMinICorr,j) = vICorr(j) - dCoeffDistI;
   
    if(bUseSparsity & t>1)
        % update - should be slower then the other updates
        G = G + 0.01 * mSparsityUpdate; %update      %  * alpha(t)
    end

    if(bNormalize)
        G(G<eps2) = 0;

        % normalize G
        Normalizer = sum(G,2);
        if(sum(Normalizer) >0)
          Normalizer(Normalizer == 0) = 10*eps;
          Normalizer = Normalizer(:,vOnes2);
          G = G./Normalizer;
        else % reset G and decrease beta0 by 50 percent 
          beta0 = (beta0 * 0.5) 
          G = sparse(rand(iNrOfPrototypes,iNrOfDataPoints));
          % normalize G
          Normalizer = sum(G,2);
          Normalizer = Normalizer(:,vOnes2);
          G = G./Normalizer;  
          'warning - decrease sparsity'
        end
    end
    % update - because G has changed
    vDistTmp = ((G*nyDis{1})*nyDis{2})*nyDis{1}';
    % slow - and intermediate matrix dim is still small
%     for i=1:iNrOfPrototypes
%         vDistTmp(i,:) =((G(i,:)*nyDis{1})*nyDis{2})*nyDis{1}'; % equivalent to (nyDis{1}*nyDis{2}*nyDis{1}') * G(i,:)' = D\alpha, order is ok
%     end        
%     
    % update statistic
    if(bActive)
        vGNormalized = sum(G,1)./iNrOfPrototypes;
        vGStd        = (2*std(vGNormalized).^2)+eps;
        vThresh      = exp(-(vGNormalized-dExpWeight).^2 ./ (vGStd));    
    end
    if(bUseSparsity)
        % log
        s = sqrt(var(G,[],2));
        dTemp = 0;
        for(p=1:iNrOfPrototypes)
              dTemp = dTemp + sum(fSparseOls(G(p,:),s(p)));
        end
        vSparsityLog(t,1) = dTemp;
        S=whos('G');
        vSparsityLog(t,2) = S.bytes;
    end
  end
  
  % redefine G using sparse - make sure to use less memory
  [i,j,s] = find(G);
  G = sparse(i,j,s,iNrOfPrototypes,iNrOfDataPoints);
  vAssignments = [];
  if(0 & mod(t,10)==0)
	[dAcc,mDO,vAssignments] = CalculateAccuracy(nyDis,G,d_class,c_class); 
    dAcc
    nnz(G)
  end
end
oLog = struct('iSkipped',{iSkipped},'vSparseness',vSparsityLog);
fprintf(1,'\n');
tElapsed=toc(tStart);
if(bNystroem)
    oResultModel = struct('K',{nyDis},'G',G,'oTrain',data,'PrototypeLabels',c_class,'vLabelMapping',vUniqueLabels,'vAssignments',vAssignments,'fKernel',pKernelFunc,'bNystroem',bNystroem,'ind',ind,'oLog',oLog,'RandState',{savedState},'ElapsedTime',tElapsed);
else
    oResultModel = struct('K',{nyDis},'G',G,'oTrain',data,'PrototypeLabels',c_class,'vLabelMapping',vUniqueLabels,'vAssignments',vAssignments,'fKernel',pKernelFunc,'bNystroem',bNystroem,'oLog',oLog,'RandState',{savedState},'ElapsedTime',tElapsed);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [dAcc,mDistancesToPrototypes,I] = CalculateAccuracy(nyDis,G, d_class,c_class)
iNrOfPrototypes = size(G,1);
iNrOfDataPoints = size(G,2);
mDistancesToPrototypes = zeros(iNrOfDataPoints,iNrOfPrototypes);
dAcc = 0.0;
vDiag    = diag(nyDis{1} * nyDis{2} * nyDis{1}')';
nyDis21  = nyDis{2}*nyDis{1}';
vDistTmp = zeros(size(G));
for i=1:iNrOfPrototypes
    vDistTmp(i,:) =((G(i,:)*nyDis{1})*nyDis{2})*nyDis{1}'; 
end
vTemp = diag(G * vDistTmp');
for i=1:iNrOfPrototypes
    mDistancesToPrototypes(:,i) = (vDiag - 2.0 * vDistTmp(i,:) + vTemp(i))';
end
[Y,I] = min(mDistancesToPrototypes,[],2);
dAcc  = sum(c_class(I) == d_class);
dAcc = dAcc/iNrOfDataPoints;

function sgd_prime = sigmoid_prime(x)

sgd_prime =  1.0/(1.0+exp(-x));
sgd_prime = sgd_prime*(1.0-sgd_prime);